/* CWinsock - Multithreaded version */

#include <string.h>
#include <ctype.h>
#include "CWinsock.h"

enum { THREAD_SINGLE, THREAD_MULTI, THREAD_CALLBACK,
       REBUILD_DONE, REBUILD_FIRST, REBUILD_SECOND };

#ifndef WINSOCK_VERSION
  #define WINSOCK_VERSION MAKEWORD(2,2)
#endif

static hostent * ResolveGeneric(char *);
static hostent * ResolveNumeric(char *);
static hostent * ResolveAlpha  (char *);

CWinsock::CWinsock()
{
   Events     = NULL;
   nEvents    = maxEvents = 0;

   hRebuild   = CreateEvent(0, 0, 0, 0);
   hKill      = CreateEvent(0, 0, 0, 0);
   hThread    = NULL;
   sRebuild   = REBUILD_DONE;
   sThread    = 0;
   bInit      = FALSE;
   
   CallBacks.Read       = DefRead;
   CallBacks.WriteRdy   = DefWriteRdy;
   CallBacks.ConnReq    = DefConnReq;
   CallBacks.Connected  = DefConnected;
   CallBacks.Close      = DefClose;
   CallBacks.Error      = DefError;

   InitializeCriticalSection(&CW_CS);
   RebuildEvents();
   ResetEvent(hRebuild);
} /* CWinsock */


CWinsock::~CWinsock()
{
   Deinit();
   CloseHandle(hRebuild);
   CloseHandle(hKill);
   free(Events);
   DeleteCriticalSection(&CW_CS);
} /* ~CWinsock */


int CWinsock::Init()
{
   if (!bInit)
   {
      if (WSAStartup(WINSOCK_VERSION, &WSData)) return 1;
      bInit = TRUE;
   }
   return InitThread();
} /* Init */


void CWinsock::Deinit()
{
   KillThread();
   
   Socks.resize(0);

   if (bInit)
   {
      WSACleanup(); // closes all sockets
      bInit = FALSE;
   }
} /* Deinit */


hostent * ResolveGeneric(char *addr)
{
   bool alpha=FALSE;
   int i, len=strlen(addr);

   for(i=0;i<len;i++)
   {
      if(isalpha(addr[i]))
      {
         alpha=TRUE;
         break;
      }
   }

   if (alpha) return ResolveAlpha(addr);
   return ResolveNumeric(addr);
} /* ResolveGeneric */


hostent * ResolveNumeric(char *addr)
{
   unsigned long address;

   address=inet_addr(addr);
   if (address == INADDR_NONE) return NULL;
   return gethostbyaddr((char *)&address,4,PF_INET);
} /* ResolveNumeric */


struct hostent * ResolveAlpha(char *addr)
{
   return gethostbyname(addr);
} /* ResolveAlpha */


int CWinsock::InitThread()
{
   SECURITY_ATTRIBUTES sa;
   DWORD id;

   if (hThread == NULL)
   {   
      sa.nLength = sizeof(SECURITY_ATTRIBUTES);
      sa.lpSecurityDescriptor = NULL;
      sa.bInheritHandle = FALSE;

      hThread = CreateThread(&sa, 0, ThreadFunc, this, 0, &id);
      if (hThread == NULL) return 1;
   }

   return 0;
} /* InitThread */

      
void CWinsock::KillThread()
{
   if (hThread)
   {
      SetEvent(hKill);
      WaitForSingleObject(hThread, INFINITE);
      hThread = NULL;
   }
} /* KillThread */


void CWinsock::RebuildEvents()
{
   int newnum = Socks.size() + 2;
   int i,j;
   CWSocket *cws;

   if (newnum > maxEvents)
   {
      Events = (HANDLE *) realloc(Events, newnum * sizeof(HANDLE));
      maxEvents = newnum;
   }
   
   nEvents = 2;
   Events[0] = hKill;
   Events[1] = hRebuild;

   for(i=2,j=0;i<newnum;j++)
   {
      cws = Socks[j];
      if (cws->hEvent)
      {
         Events[i++] = cws->hEvent;
         nEvents++;
      }
   }

} /* RebuildEvents */


int CWinsock::Open(CWSocket *cws, CWType type, bool blocking)
{
   cws->Type = type;

   cws->myOpts.nonblocking = !blocking;
   if (ioctlsocket(cws->hSock,FIONBIO,(u_long *)&cws->myOpts.nonblocking)) return 1;
   if (cws->GetOpts(&cws->myOpts)) return 1;

   if (blocking) return 0;

   cws->hEvent = WSACreateEvent();
   if (!cws->hEvent) return 1;
   if (WSAEventSelect(cws->hSock, cws->hEvent,
                  FD_READ | FD_WRITE | FD_ACCEPT | FD_CONNECT | FD_CLOSE)) return 1;
   
   if (sRebuild != REBUILD_DONE)
   {
      while(sThread != THREAD_CALLBACK && sRebuild != REBUILD_DONE); // cant be rebuilding at the same time
   }
   sRebuild = REBUILD_FIRST;
   SetEvent(hRebuild);
   EnterCriticalSection(&CW_CS);
   Socks.push_back(cws);
   RebuildEvents();
   SetEvent(hRebuild);
   sRebuild = REBUILD_SECOND;
   LeaveCriticalSection(&CW_CS);

   return 0;
} /* Open */


int CWinsock::Close(CWSocket *cws)
{
   return Close(cws, cws->myOpts.lingerset.l_onoff,cws->myOpts.lingerset.l_linger);
} /* Close(CWSocket *) */


int CWinsock::Close(CWSocket *cws, BOOL lingon, int timeout)
{
   linger ling;
   HANDLE handle;

   ling.l_onoff=(unsigned short)lingon;
   ling.l_linger=(unsigned short)timeout;

   if (cws->hEvent)
   {
      WSAEventSelect(cws->hSock, NULL, 0);
      if (sRebuild != REBUILD_DONE)
      {
         while(sThread != THREAD_CALLBACK && sRebuild != REBUILD_DONE); // cant be rebuilding at the same time
      }
      sRebuild = REBUILD_FIRST;
      SetEvent(hRebuild);
      EnterCriticalSection(&CW_CS);
      std::vector<CWSocket *>::iterator i = Socks.begin();
      for(;i != Socks.end();i++)
      {
         if(*i == cws)
         {
            Socks.erase(i);
            break;
         }
      }
      handle = cws->hEvent;
      cws->hEvent = NULL;
      RebuildEvents();
      SetEvent(hRebuild);
      sRebuild = REBUILD_SECOND;
      LeaveCriticalSection(&CW_CS);
      WSACloseEvent(handle);
   }

   if (cws->hSock != INVALID_SOCKET)
   {
      setsockopt(cws->hSock,SOL_SOCKET,SO_DONTLINGER,(char const *)&lingon,sizeof(bool));
      setsockopt(cws->hSock,SOL_SOCKET,SO_LINGER,(char const *)&ling,sizeof(linger));
      closesocket(cws->hSock);
      cws->hSock=INVALID_SOCKET;
   }

   return 0;
} /* Close(CWSocket *, bool lingon, int timeout) */


int CWinsock::OpenSocket(CWSocket *cws)
{
   CloseSocket(cws, 0, 0);

   cws->hSock        = INVALID_SOCKET;
   cws->hEvent       = NULL;
   cws->RemAddr[0]   = 0;
   cws->RemPort      = 0;
   cws->Parent       = this;

   return 0;
} /* OpenSocket(CWSocket *) */


int CWinsock::OpenSocket(CWSocket *cws, CWType type, bool blocking)
{
   if (OpenSocket(cws)) return 1;

   cws->hSock = socket(PF_INET, type, 0);
   if (!cws->hSock) return 1;

   return Open(cws, type, blocking);
} /* OpenSocket(CWSocket *, CWType, bool) */


int CWinsock::OpenSocket(CWSocket *cws, SOCKET s, CWType type, bool blocking)
{
   if (OpenSocket(cws) || s == INVALID_SOCKET) return 1;
   cws->hSock = s;
   return Open(cws, type, blocking);
} /* OpenSocket(CWSocket *, SOCKET, bool) */


int CWinsock::CloseSocket(CWSocket *cws)
{
   return Close(cws);
} /* CloseSocket(CWSocket *) */


int CWinsock::CloseSocket(CWSocket *cws, bool lingon, int timeout)
{
   return Close(cws, lingon, timeout);
} /* CloseSocket(CWSocket *) */


int CWSocket::Bind(int port)
{
   SOCKADDR_IN sin;
   sin.sin_family=AF_INET;
   sin.sin_port=htons((unsigned short)port);
   sin.sin_addr.s_addr=ADDR_ANY;

   return bind(hSock, (LPSOCKADDR)&sin, sizeof(sin));
} /* Bind(CWSocket *, int port) */


int CWSocket::Bind(char *addr, int port)
{
   hostent *res;
   SOCKADDR_IN sin;

   res = ResolveGeneric(addr);
   if (res == NULL) return 1;

   sin.sin_family=AF_INET;
   sin.sin_port=htons((unsigned short)port);
   sin.sin_addr.s_addr=*(u_long *)res->h_addr_list[0];

   return bind(hSock,(LPSOCKADDR)&sin,sizeof(sin));
} /* Bind(CWSocket *, char *addr, int port) */


int CWSocket::Listen(int log)
{
   return ::listen(hSock, log);
} /* Listen */


int CWSocket::Accept(CWSocket *newsock)
{
   SOCKET s;

   s = ::accept(hSock, 0, 0);
   if (s == INVALID_SOCKET) return 1;

   return Parent->OpenSocket(newsock, s, Type, !myOpts.nonblocking);
} /* Accept */


int CWSocket::Connect(char *addr, int port)
{
   hostent    *host;
   sockaddr_in sa;

   host = ResolveGeneric(addr);
   if (!host) return 1;

   sa.sin_addr.s_addr = *(u_long*)host->h_addr_list[0];
   sa.sin_family = AF_INET;
   sa.sin_port   = htons((unsigned short)port);

   if (Parent->OpenSocket(this, Type, !myOpts.nonblocking)) return 1;
   
   return connect(hSock, (sockaddr *)&sa, sizeof(sa));
} /* Connect */


int CWSocket::ShutDown()
{
   return shutdown(hSock, SD_SEND);
} /* ShutDown */


int CWSocket::Send(char *buf, int len)
{
   return send(hSock, buf, len, 0);
} /* Send */


int CWSocket::Recv(char *buf, int len, int flags)
{
   return recv(hSock, buf, len, flags);
} /* Recv */


int CWSocket::SendTo(char *buf, int len, CWIPAddress *addr)
{
   return sendto(hSock, buf, len, 0, (sockaddr *)addr->GetIPSockAddr(), sizeof(sockaddr_in));
} /* SendTo */


int CWSocket::RecvFrom(char *buf, int len, int flags, CWIPAddress *addr)
{
   int alen = sizeof(sockaddr_in);
   return recvfrom(hSock, buf, len, flags, (sockaddr *)addr->GetIPSockAddr(), &alen);
} /* RecvFrom */


int CWSocket::SetOpts(CWSockOpts *opts)
{
   int ret=0, len=sizeof(u_long);

   myOpts=*opts;

   ret|=setsockopt(hSock,SOL_SOCKET,SO_DEBUG,(char const *)&(opts->debug),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_DONTLINGER,(char const *)&(opts->dontlinger),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_DONTROUTE,(char const *)&(opts->dontroute),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_KEEPALIVE,(char const *)&(opts->keepalive),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_LINGER,(char const *)&(opts->lingerset),sizeof(struct linger));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_OOBINLINE,(char const *)&(opts->oobinline),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_REUSEADDR,(char const *)&(opts->reuseaddr),sizeof(bool));
   ret|=setsockopt(hSock,IPPROTO_TCP,TCP_NODELAY,(char const *)&(opts->nodelay),sizeof(bool));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_RCVBUF,(char const *)&(opts->rcvbuf),sizeof(int));
   ret|=setsockopt(hSock,SOL_SOCKET,SO_SNDBUF,(char const *)&(opts->sndbuf),sizeof(int));
   ret|=ioctlsocket(hSock,FIONBIO,(u_long FAR *)opts->nonblocking);

   // re-read buffer settings to get actual values obtained
   ret|=getsockopt(hSock,SOL_SOCKET,SO_RCVBUF,(char *)&(myOpts.rcvbuf),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_SNDBUF,(char *)&(myOpts.sndbuf),&len);

   return ret;
} /* SetOpts */


int CWSocket::GetOpts(CWSockOpts *opts)
{
   int ret = 0;
   int len = sizeof(int);
   CWSockOpts temp;

   ret|=getsockopt(hSock,SOL_SOCKET,SO_DEBUG,(char *)&(temp.debug),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_DONTLINGER,(char *)&(temp.dontlinger),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_DONTROUTE,(char *)&(temp.dontroute),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_KEEPALIVE,(char *)&(temp.keepalive),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_OOBINLINE,(char *)&(temp.oobinline),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_REUSEADDR,(char *)&(temp.reuseaddr),&len);
   ret|=getsockopt(hSock,IPPROTO_TCP,TCP_NODELAY,(char *)&(temp.nodelay),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_RCVBUF,(char *)&(temp.rcvbuf),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_SNDBUF,(char *)&(temp.sndbuf),&len);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_MAX_MSG_SIZE,(char *)&(temp.maxsize),&len);
   len = sizeof(linger);
   ret|=getsockopt(hSock,SOL_SOCKET,SO_LINGER,(char *)&(temp.lingerset),&len);
   temp.nonblocking = myOpts.nonblocking;

   myOpts = *opts = temp;

   return ret;
} /* GetOpts */


SOCKET CWSocket::GetSocket() const
{
   return hSock;
} /* GetSocket */


int CWSocket::DataWaiting() const
{
   unsigned long len;

   if (ioctlsocket(hSock, FIONREAD, &len)) return -1;
   return len;
} /* DataWaiting */


char const * CWSocket::GetPeerIP() const
{
   static char addr[16];
   sockaddr_in sad;
   int len = sizeof(sad);
   int ret = getpeername(hSock, (sockaddr *)&sad, &len);
   if(ret) return NULL;

   wsprintf(addr,"%d.%d.%d.%d\0",
            sad.sin_addr.S_un.S_un_b.s_b1, sad.sin_addr.S_un.S_un_b.s_b2,
            sad.sin_addr.S_un.S_un_b.s_b3, sad.sin_addr.S_un.S_un_b.s_b4);
   return addr;
} /* GetPeerIP */


int CWSocket::GetPeerPort() const
{
   sockaddr_in sad;
   int len = sizeof(sad);
   if(getpeername(hSock, (sockaddr *)&sad, &len)) return -1;
   return sad.sin_port;
} /* GetPeerPort */


void CWinsock::SetCallbacks(CWCalls *cwc)
{
   CallBacks = *cwc;
} /* SetCallbacks */


void CWinsock::GetCallbacks(CWCalls *cwc) const
{
   *cwc = CallBacks;
} /* GetCallbacks */


int CWinsock::GetLocalInfo(char *addr, int len) const
{
   return gethostname(addr, len);
} /* GetLocalInfo */


int CWinsock::LastError() const
{
   return WSAGetLastError();
} /* LastError */


int CWinsock::GetWSData(WSADATA *wsd) const
{
   *wsd = WSData;

   return 0;
} /* GetWSData */


int CWinsock::DefRead(CWSocket *cws)
{
   char buf[512];

   // read and discard up to 512 bytes (apps should use Recv, not recv)
   return cws->Recv(buf, 512);
} /* DefRead */


int CWinsock::DefWriteRdy(CWSocket *cws)
{
   return 0;
} /* DefWriteRdy */


int CWinsock::DefConnReq(CWSocket *cws)
{
   return 0;
} /* DefConnReq */


int CWinsock::DefConnected(CWSocket *cws)
{
   return 0;
} /* DefConnected */


int CWinsock::DefClose(CWSocket *cws)
{
   return 0;
} /* DefClose */


int CWinsock::DefError(CWSocket *cws, int cwe)
{
   return 0;
} /* DefError */


DWORD WINAPI CWinsock::ThreadFunc(LPVOID cls)
{
   DWORD event;
   CWSocket *cws;
   CWinsock *Me = (CWinsock *)cls;
   WSANETWORKEVENTS wne;
   int Err=0;

   while(TRUE)
   {
      if (Me->sRebuild == REBUILD_FIRST)
      {
         Me->sThread = THREAD_SINGLE;
         WaitForSingleObject(Me->hRebuild, INFINITE);
         Me->sRebuild = REBUILD_DONE;
         Sleep(0);
      }
      EnterCriticalSection(&Me->CW_CS);
      Me->sThread = THREAD_MULTI;
      if (Me->sRebuild == REBUILD_SECOND) Me->sRebuild = REBUILD_DONE;
      event = WaitForMultipleObjects(Me->nEvents, Me->Events, FALSE, INFINITE);

      switch(event)
      {
         case WAIT_OBJECT_0: // hKill
            LeaveCriticalSection(&Me->CW_CS);
            ExitThread(0);
            break;
         
         case WAIT_OBJECT_0 + 1: // hRebuild
            LeaveCriticalSection(&Me->CW_CS);
            break;

         default:
            Me->sThread = THREAD_CALLBACK;
            cws = Me->Socks[event - WAIT_OBJECT_0 - 2];
            if (WSAEnumNetworkEvents(cws->hSock, cws->hEvent, &wne)) break;

            if (wne.lNetworkEvents & FD_READ)
            {
               if (wne.iErrorCode[FD_READ_BIT]) Err |= CWE_READ;
               else if (Me->CallBacks.Read) Me->CallBacks.Read(cws);
            }
            if (wne.lNetworkEvents & FD_WRITE)
            {
               if (wne.iErrorCode[FD_WRITE_BIT]) Err |= CWE_WRITE;
               else if (Me->CallBacks.WriteRdy) Me->CallBacks.WriteRdy(cws);
            }
            if (wne.lNetworkEvents & FD_ACCEPT)
            {
               if (wne.iErrorCode[FD_ACCEPT_BIT]) Err |= CWE_ACCEPT;
               else if (Me->CallBacks.ConnReq) Me->CallBacks.ConnReq(cws);
            }
            if (wne.lNetworkEvents & FD_CONNECT)
            {
               if (wne.iErrorCode[FD_CONNECT_BIT]) Err |= CWE_CONNECT;
               else if (Me->CallBacks.Connected) Me->CallBacks.Connected(cws);
            }
            if (wne.lNetworkEvents & FD_CLOSE)
            {
               if (wne.iErrorCode[FD_CLOSE_BIT]) Err |= CWE_CLOSE;
               else if (Me->CallBacks.Close) Me->CallBacks.Close(cws);
            }
            if (Me->CallBacks.Error && Err) Me->CallBacks.Error(cws, Err);

            LeaveCriticalSection(&Me->CW_CS);
            break;
      }
   }

   return 0; // will never get here
} /* ThreadFunc */


/* ------- FOR CWIPADDRESS ------- */

CWIPAddress::CWIPAddress()
{
   ZeroMemory(&Address, sizeof(Address));
   Address.sin_family = AF_INET;
}


CWIPAddress::CWIPAddress(char *addr, int port)
{
   CWIPAddress(); // call default contructor
   SetIPAddress(addr, port);
}


CWIPAddress::CWIPAddress(DWORD addr, int port)
{
   CWIPAddress(); // call default contructor
   SetIPAddress(addr, port);
}


CWIPAddress::CWIPAddress(sockaddr_in *addr)
{
   CWIPAddress(); // call default contructor
   SetIPAddress(addr);
}


int CWIPAddress::SetIPAddress(char *addr, int port)
{
   hostent *host = ResolveGeneric(addr);
   if (!host) return 1;

   return SetIPAddress(*(DWORD*)host->h_addr_list[0], port);
}


int CWIPAddress::SetIPAddress(DWORD addr, int port)
{
   Address.sin_addr.s_addr=(u_long)addr;
   Address.sin_port=htons((unsigned short)port);
   return 0;
}


int CWIPAddress::SetIPAddress(sockaddr_in *addr)
{
   Address = *addr;
   return 0;
}


char * CWIPAddress::GetIPString()
{
   static char straddr[16];
   char *addr = (char *)&Address.sin_addr.s_addr;

   wsprintf(straddr,"%d.%d.%d.%d\0",addr[0],addr[1],addr[2],addr[3]);

   return straddr;
}


DWORD CWIPAddress::GetIPLong()
{
   return (DWORD)(Address.sin_addr.s_addr);
}


sockaddr_in * CWIPAddress::GetIPSockAddr()
{
   return &Address;
}
